from tracardi.domain.value_object.storage_info import StorageInfo

from tracardi.domain.entity import Entity
from datetime import datetime


class ApiInstance(Entity):
    ip: str = None
    timestamp: datetime = datetime.utcnow()

    # Persistence

    @staticmethod
    def storage_info() -> StorageInfo:
        return StorageInfo(
            'api-instance',
            ApiInstance
        )


# import elasticsearch
# import pytest
#
# import tracardi
# from .mocks.mock_storage import MockStorageCrud
# from tracardi.domain.entity import Entity
# from tracardi.domain.profile import Profile
# from tracardi.domain.session import Session
# from tracardi.domain.payload.tracker_payload import TrackerPayload
#
#
# @pytest.mark.asyncio
# async def test_get_context_session_exists_profile_not_exists(mocker):
#
#     mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
#     mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
#     mocker.patch('tracardi.domain.entity.crud', tracardi.tests.unit.mocks.mock_storage)
#     mocker.patch('tracardi.domain.profile.StorageCrud', MockStorageCrud)
#
#     # Session exists and has profile equal to 1
#     tracker_payload = TrackerPayload(
#         session=Entity(id="1"),
#         profile=None,
#         source=Entity(id="scope")
#     )
#
#     profile, session = await tracker_payload.get_profile_and_session()
#
#     assert isinstance(profile, Profile)
#     assert isinstance(session, Session)
#
#     assert profile.id == "1"
#     assert session.id == "1"
#
#
# @pytest.mark.asyncio
# async def test_get_context_session_exists_profile_exists_conflict(mocker):
#     mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
#     mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
#     mocker.patch('tracardi.domain.entity.crud', tracardi.tests.unit.mocks.mock_storage)
#     mocker.patch('tracardi.domain.profile.StorageCrud', MockStorageCrud)
#
#     # Session exists and has profile equal to 1
#     tracker_payload = TrackerPayload(
#         session=Entity(id="1"),
#         profile=Entity(id="2"),
#         source=Entity(id="scope")
#     )
#
#     profile, session = await tracker_payload.get_profile_and_session()
#
#     assert isinstance(profile, Profile)
#     assert isinstance(session, Session)
#
#     # profile from session is most important
#     assert profile.id == "1"
#     assert session.id == "1"
#
#
# @pytest.mark.asyncio
# async def test_get_context_session_not_exists_profile_exists(mocker):
#     mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
#     mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
#     mocker.patch('tracardi.domain.entity.crud', tracardi.tests.unit.mocks.mock_storage)
#     mocker.patch('tracardi.domain.profile.StorageCrud', MockStorageCrud)
#
#     # Session 2 does not exist
#     # Profile 2 exists
#     tracker_payload = TrackerPayload(
#         session=Entity(id="generated_not_exists-01"),
#         profile=Entity(id="2"),
#         source=Entity(id="scope")
#     )
#
#     profile, session = await tracker_payload.get_profile_and_session()
#
#     assert isinstance(profile, Profile)
#     assert isinstance(session, Session)
#
#     # Do not generate profile copy it from param
#     assert session.id == "generated_not_exists-01"
#     assert profile.id == "2"
#
#
# @pytest.mark.asyncio
# async def test_get_context_session_not_exists_profile_not_exists(mocker):
#     mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
#     mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
#     mocker.patch('tracardi.domain.entity.crud', tracardi.tests.unit.mocks.mock_storage)
#     mocker.patch('tracardi.domain.profile.StorageCrud', MockStorageCrud)
#
#     tracker_payload = TrackerPayload(
#         session=Entity(id="generated_not_exists-02"),
#         profile=None,
#         source=Entity(id="scope")
#     )
#
#     profile, session = await tracker_payload.get_profile_and_session()
#
#     assert isinstance(profile, Profile)
#     assert isinstance(session, Session)
#
#     # Profile generated
#     assert profile.id != "2"
#     # Session generated
#     assert session.id == "generated_not_exists-02"
#
#
# @pytest.mark.asyncio
# async def test_load_session_with_error(mocker):
#     def exc(*args):
#         raise elasticsearch.exceptions.ElasticsearchException("test")
#
#     mocker.patch('tracardi.domain.session.StorageCrud.load', exc)
#
#     try:
#         tracker_payload = TrackerPayload(
#             session=Entity(id="non_existent"),
#             profile=None,
#             source=Entity(id="scope")
#         )
#
#         profile, session = await tracker_payload.get_profile_and_session()
#
#     except elasticsearch.exceptions.ElasticsearchException as e:
#         assert str(e) == "test"



import elasticsearch
from typing import Union
from tracardi.service.storage.elastic_storage import ElasticStorage
from tracardi.service.storage.persistence_service import PersistenceService
from tracardi.domain.entity import Entity
from pydantic import BaseModel
from tracardi.domain.agg_result import AggResult
from tracardi.exceptions.exception import TracardiException, StorageException
from tracardi.domain.storage_result import StorageResult
from tracardi.domain.value_object.bulk_insert_result import BulkInsertResult
import tracardi.domain.entity as domain


class BaseStorageCrud:

    def __init__(self, index, entity, exclude=None, exclude_unset=False):
        self.exclude_unset = exclude_unset
        self.exclude = exclude
        self.index = index
        self.entity = entity

    async def delete(self) -> dict:
        service = self._get_storage_service()
        return await service.delete(self.entity.id)

    def _get_storage_service(self):
        return storage(self.index)


class EntityStorageCrud(BaseStorageCrud):

    async def load(self, domain_class_ref=None):
        service = self._get_storage_service()
        data = await service.load(self.entity.id)

        if data:
            if domain_class_ref is None:
                return domain.Entity(**data)
            return domain_class_ref(**data)

        return None

    async def load_by(self, field: str, value: str, limit: int = 100) -> StorageResult:
        service = self._get_storage_service()
        return await service.load_by(field, value, limit)

    async def delete_by(self, field, value) -> dict:
        service = self._get_storage_service()
        return await service.delete_by(field, value)

    async def save(self, row=None):
        if row is None:
            row = {}
        row['id'] = self.entity.id
        service = self._get_storage_service()
        return await service.upsert(row)


class StorageCrud(BaseStorageCrud):

    def __init__(self, index, domain_class_ref, entity, exclude=None, exclude_unset=False):
        super().__init__(index, entity, exclude, exclude_unset)
        self.domain_class_ref = domain_class_ref

    async def load(self):
        service = self._get_storage_service()
        data = await service.load(self.entity.id)

        if data:
            return self.domain_class_ref(**data)

        return None

    async def save(self) -> BulkInsertResult:
        service = self._get_storage_service()
        return await service.upsert(self.entity.dict(exclude_unset=self.exclude_unset, exclude=self.exclude))


class CollectionCrud:

    def __init__(self, index, payload):
        self.payload = payload
        self.index = index
        self.storage = storage(self.index)

    async def save(self) -> BulkInsertResult:
        try:
            if not isinstance(self.payload, list):
                raise TracardiException("CollectionCrud dave payload must be list.")
            data = [p.dict() for p in self.payload if isinstance(p, BaseModel)]
            return await self.storage.upsert(data)

        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def load(self) -> StorageResult:
        try:

            return await self.storage.load_all()

        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def uniq_field_value(self, field) -> AggResult:
        try:
            query = {
                "size": "0",
                "aggs": {
                    "uniq": {
                        "terms": {
                            "field": field
                        }
                    }
                }
            }
            return AggResult('uniq', await self.storage.query(query), return_counts=False)
        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))


class StorageFor:

    def __init__(self, instance: Entity):
        self.instance = instance
        self.storage_info = instance.storage_info()

    @staticmethod
    def crud(index, class_type):
        return EntityStorageCrud(index, class_type)

    def index(self, index=None) -> Union[EntityStorageCrud, StorageCrud]:
        if self.storage_info is None:
            if index is None:
                raise ValueError("When loading entity of type `{}` you must provide index of that entity.".format(
                    type(self.instance)
                ))
            return EntityStorageCrud(index, entity=self.instance)
        else:
            return StorageCrud(
                self.storage_info.index,
                self.storage_info.domain_class_ref,
                entity=self.instance,
                exclude=self.storage_info.exclude,
                exclude_unset=self.storage_info.exclude_unset
            )


class StorageForBulk:

    def __init__(self, list_of_objects=None):
        if list_of_objects is None:
            list_of_objects = []
        self.collection = list_of_objects

    def index(self, index) -> CollectionCrud:
        return CollectionCrud(index, self.collection)


def storage(index) -> PersistenceService:
    return PersistenceService(ElasticStorage(index_key=index))



from datetime import datetime
from typing import Optional, List, Tuple, Any
from pydantic import BaseModel
from tracardi.domain.event import Event

from ..entity import Entity
from ..payload.event_payload import EventPayload
from ..profile import Profile
from ..session import Session
from ..metadata import Metadata
from ..time import Time


class TrackerPayload(BaseModel):
    source: Entity
    session: Entity

    metadata: Optional[Metadata]
    profile: Optional[Entity] = None
    context: Optional[dict] = {}
    properties: Optional[dict] = {}
    events: List[EventPayload] = []
    options: Optional[dict] = {}

    def __init__(self, **data: Any):
        data['metadata'] = Metadata(
            time=Time(
                insert=datetime.utcnow()
            ))
        super().__init__(**data)

    def get_events(self, session: Session, profile: Profile) -> List[Event]:
        _events = []
        if self.events:
            for event in self.events:  # type: EventPayload
                _event = event.to_event(self.metadata, self.source, session, profile, event.options)
                _events.append(_event)
        return _events

    def return_profile(self):
        return self.options and "profile" in self.options and self.options['profile'] is True

    def is_disabled(self, key):
        return key in self.options and self.options[key] is False

    # async def save_collected_data(self, profile: Profile, session: Session, events: list) -> Tuple[Profile, Session, list, CollectResult]:
    #
    #     # Get profile, session objects
    #     # profile, session = await self.get_profile_and_session()
    #     # events = self.get_events(session, profile)
    #
    #     # Presistence
    #
    #     save_session = False if self.is_disabled('saveSession') else True
    #     save_events = False if self.is_disabled('saveEvents') else True
    #     save_profile = True
    #
    #     # Save profile
    #     if save_profile and profile.operation.new:
    #         profile_result = await StorageFor(profile).index().save()
    #     else:
    #         profile_result = BulkInsertResult()
    #
    #     # Save session
    #     if save_session and (session.operation.new or session.profile is None or session.profile.id != profile.id):
    #         # save only profile Entity
    #         session.profile = Entity(id=profile.id)
    #         session_result = await StorageFor(session).index().save()
    #     else:
    #         session_result = BulkInsertResult()
    #
    #     # Save events
    #     if save_events:
    #         events_to_save = [event for event in events if event.is_persistent()]
    #         event_result = await StorageForBulk(events_to_save).index('event').save()
    #         event_result = SaveResult(**event_result.dict())
    #
    #         # Add event types
    #         for event in events:
    #             event_result.types.append(event.type)
    #     else:
    #         event_result = BulkInsertResult()
    #
    #     result = {
    #         "session": session_result,
    #         "events": event_result,
    #         "profile": profile_result
    #
    #     }
    #     return profile, session, events, CollectResult(**result)

    async def get_profile_and_session(self, session: Session, load_merged_profile) -> Tuple[Profile, Session]:

        """
        Returns session. Creates profile if it does not exist.If it exists connects session with profile.
        """

        is_new_profile = False
        is_new_session = False

        if session is None:  # loaded session is empty

            session = Session(id=self.session.id)
            is_new_session = True

            # Bind profile

            if self.profile is None:

                # Create empty default profile generate profile_id
                profile = Profile.new()

                # Create new profile
                is_new_profile = True

            else:

                # Id exists load profile from storage
                profile = await load_merged_profile(id=self.profile.id)  # type: Profile

                if profile is None:
                    # Profile id delivered but profile does not exist in storage.
                    # Id was forged

                    profile = Profile.new()

                    # Create new profile
                    is_new_profile = True

        else:

            # There is session. Copy profile id form session to profile

            if session.profile is not None:
                # Loaded session has profile

                # Load profile based on profile id saved in session
                profile = await load_merged_profile(id=session.profile.id)  # type: Profile

                if session.profile.id != profile.id:
                    # Profile in session id has been merged. Change profile in session.

                    session.profile.id = profile.id
                    session.metadata.time = Time(insert=datetime.utcnow())

                    is_new_session = True

                # profile = await profile.storage().load()  # type: Profile

            else:
                # Corrupted session

                profile = None

            if profile is None:
                # Id exists but profile not exist in storage.

                profile = Profile.new()

                # Create new profile
                is_new_profile = True

        session.context = self.context
        session.properties = self.properties
        session.operation.new = is_new_session

        profile.operation.new = is_new_profile

        return profile, session

    # async def process(self, profile, session, events) -> Tuple[dict, Profile]:
    #
    #     # profile, session, events, collect_result = await self.save_collected_data(profile, session, events)
    #
    #     rules_engine = RulesEngine(
    #         session,
    #         profile,
    #         events)
    #
    #     debug_info_by_event_type_and_rule_name, segmentation_result, console_log = await rules_engine.execute(
    #         self.source.id
    #     )
    #
    #     # Prepare response
    #     result = {}
    #
    #     # Debugging
    #     # todo save result to different index
    #     if not tracardi.track_debug and not self.is_disabled('debugger'):
    #         debug_result = TrackerPayloadResult(**collect_result.dict())
    #         debug_result = debug_result.dict()
    #         debug_result['execution'] = debug_info_by_event_type_and_rule_name
    #         debug_result['segmentation'] = segmentation_result
    #         debug_result['logs'] = console_log
    #         result['debugging'] = debug_result
    #
    #     # Profile
    #     if self.return_profile():
    #         result["profile"] = profile.dict(
    #             exclude={
    #                 "traits": {"private": ...},
    #                 "pii": ...,
    #                 "operation": ...
    #             })
    #     else:
    #         result["profile"] = profile.dict(include={"id": ...})
    #
    #     return result, profile



from tracardi.domain.entity import Entity
from tracardi.domain.resource import ResourceRecord, Resource
from tracardi.service.storage.factory import StorageFor


async def read_source(resource_id: str) -> Resource:
    """
    Reads source by source id
    :param resource_id: str
    :return source: Source
    """

    resource_config_record = await StorageFor(Entity(id=resource_id)). \
        index('resource'). \
        load(ResourceRecord)  # type: ResourceRecord

    if resource_config_record is None:
        raise ValueError('Resource id {} does not exist.'.format(resource_id))

    return resource_config_record.decode()



from tracardi.domain.value_object.bulk_insert_result import BulkInsertResult
from tracardi.domain.record.event_debug_record import EventDebugRecord
from tracardi.service.storage.factory import StorageForBulk


async def save_debug_info(debug_info_by_event_type_and_rule_name) -> BulkInsertResult:
    record = []
    for debug_info_record in EventDebugRecord.encode(
            debug_info_by_event_type_and_rule_name):  # type: EventDebugRecord
        record.append(debug_info_record)

    # Save in debug index
    return await StorageForBulk(record).index("debug-info").save()



from typing import List
from tracardi.domain.entity import Entity
from tracardi.domain.profile import Profile
from tracardi.service.storage.factory import StorageFor, storage


async def load_merged_profile(id: str) -> Profile:
    """
    Loads current profile. If profile was merged then it loads merged profile.
    """

    entity = Entity(id=id)
    profile = await StorageFor(entity).index('profile').load(Profile)  # type: Profile
    if profile is not None and profile.mergedWith is not None:
        # Has merged profile
        profile = await load_merged_profile(profile.mergedWith)
    return profile


async def load_profiles_to_merge(merge_key_values: List[tuple], limit=1000) -> List[Profile]:
    profiles = await storage('profile').load_by_values(merge_key_values, limit)
    return [Profile(**profile) for profile in profiles]



from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.action_runner import ActionRunner
from tracardi_plugin_sdk.domain.result import Result


class ReadSessionAction(ActionRunner):

    def __init__(self, **kwargs):
        pass

    async def run(self, payload):
        return Result(port="payload", value=self.session.dict())


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.read_session_action',
            className='ReadSessionAction',
            inputs=["payload"],
            outputs=['payload'],
            init=None,
            manual="read_session_action",
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Read session',
            desc='Loads session data and returns it.',
            type='flowNode',
            width=200,
            height=100,
            icon='json',
            group=["Read"]
        )
    )



# from typing import Optional, List
#
# from pydantic import BaseModel
#
# from tracardi.domain.entity import Entity
# from tracardi.domain.value_object.storage_info import StorageInfo
# from tracardi.service.storage.crud import StorageCrud
#
#
# class ResponseItem(BaseModel):
#     index: str
#     slices: Optional[List[str]] = None
#
#
# class Response(Entity):
#     data: List[ResponseItem]
#
#     def storage(self) -> StorageCrud:
#         return StorageCrud("response", Response, entity=self)
#
#     @staticmethod
#     def storage_info() -> StorageInfo:
#         return StorageInfo(
#             'response',
#             Response
#         )



# import logging
#
# from tracardi.event_server.service.persistence_service import PersistenceService
# from tracardi.service.storage.collection_crud import CollectionCrud
# from tracardi.service.storage.elastic_storage import ElasticStorage
#
#
# rules_logger = logging.getLogger('Flows')
#
#
# class Flows(list):
#
#     def bulk(self) -> CollectionCrud:
#         return CollectionCrud("flow", self)
#



from datetime import datetime
from typing import Optional, Any
from uuid import uuid4

from .entity import Entity
from .metadata import Metadata
from .time import Time
from .value_object.operation import Operation
from .value_object.storage_info import StorageInfo


class Session(Entity):
    metadata: Optional[Metadata]
    operation: Operation = Operation()
    profile: Optional[Entity] = None
    context: Optional[dict] = {}
    properties: Optional[dict] = {}

    def __init__(self, **data: Any):
        data['metadata'] = Metadata(
            time=Time(
                insert=datetime.utcnow()
            ))
        super().__init__(**data)

    def replace(self, session):
        self.metadata = session.metadata
        self.profile = session.profile
        self.context = session.context
        self.id = session.id
        self.properties = session.properties
        self.operation = session.operation

    @staticmethod
    def storage_info() -> StorageInfo:
        return StorageInfo(
            'session',
            Session
        )

    @staticmethod
    def new() -> 'Session':
        return Session(id=str(uuid4()))



# import logging
# from tracardi.event_server.service.persistence_service import PersistenceService
# from tracardi.service.storage.collection_crud import CollectionCrud
# from tracardi.service.storage.elastic_storage import ElasticStorage
#
#
# _logger = logging.getLogger('FlowActionPlugins')
#
#
# class FlowActionPlugins(list):
#
#     def bulk(self) -> CollectionCrud:
#         return CollectionCrud("action", self)



import asyncio
from asyncio import Task
from collections import defaultdict
from time import time
from typing import Dict, List, Tuple
from pydantic import ValidationError
from tracardi.domain.event import Event

from tracardi_graph_runner.domain.debug_info import DebugInfo
from tracardi_graph_runner.domain.error_debug_info import ErrorDebugInfo
from tracardi_graph_runner.domain.debug_info import FlowDebugInfo
from tracardi_graph_runner.domain.flow_history import FlowHistory
from tracardi_graph_runner.domain.work_flow import WorkFlow
from tracardi_plugin_sdk.domain.console import Log
from ..domain.console import Console
from ..domain.entity import Entity
from ..domain.profile import Profile
from ..domain.session import Session
from ..domain.rule import Rule


class RulesEngine:

    def __init__(self,
                 session: Session,
                 profile: Profile,
                 events_rules: List[Tuple[Task, Event]],
                 console_log=None
                 ):

        if console_log is None:
            console_log = []

        self.console_log = console_log
        self.session = session
        self.profile = profile
        self.events_rules = events_rules

    async def invoke(self, load_flow_callable, source_id=None) -> Tuple[Dict[str, List[Dict[str, DebugInfo]]], list, list]:

        flow_task_store = defaultdict(list)
        debug_info_by_event_type_and_rule_name = defaultdict(list)

        for rules_loading_task, event in self.events_rules:

            # Loads rules only for event.type
            rules = await rules_loading_task

            for rule in rules:

                try:
                    rule = Rule(**rule)
                except ValidationError as e:
                    print(str(e))
                    # todo log error and disable rule
                    continue

                if not rule.enabled:
                    continue

                try:
                    flow = await load_flow_callable(rule.flow.id)
                except Exception as e:
                    # This is empty DebugInfo without nodes
                    debug_info = DebugInfo(
                        timestamp=time(),
                        flow=FlowDebugInfo(
                            id=rule.flow.id,
                            error=[ErrorDebugInfo(msg=str(e), file=__file__, line=103)]
                        ),
                        event=Entity(id=event.id)
                    )
                    debug_info_by_event_type_and_rule_name[event.type].append({rule.name: debug_info})
                    continue

                if not flow.enabled:
                    continue

                # Validates rule source. Type was verified before

                if source_id:
                    if source_id == event.source.id:
                        # todo FlowHistory is empty
                        workflow = WorkFlow(
                            FlowHistory(history=[]),
                            self.session,
                            self.profile,
                            event
                        )
                        flow_task = asyncio.create_task(workflow.invoke(flow, debug=False))
                        flow_task_store[event.type].append((rule.flow.id, event.id, rule.name, flow_task))
                else:
                    # todo FlowHistory is empty
                    workflow = WorkFlow(
                        FlowHistory(history=[]),
                        self.session,
                        self.profile,
                        event
                    )
                    flow_task = asyncio.create_task(workflow.invoke(flow, debug=False))
                    flow_task_store[event.type].append((rule.flow.id, event.id, rule.name, flow_task))

        # Run and report async

        for event_type, tasks in flow_task_store.items():
            for flow_id, event_id, rule_name, task in tasks:
                try:
                    debug_info, log_list = await task  # type: DebugInfo, List[Log]

                    # Store logs in one console log
                    for log in log_list:  # type: Log
                        console = Console(
                            origin="node",
                            event_id=event_id,
                            flow_id=flow_id,
                            module=log.module,
                            class_name=log.class_name,
                            type=log.type,
                            message=log.message
                        )
                        self.console_log.append(console)

                except Exception as e:
                    # todo log error

                    console = Console(
                        origin="workflow",
                        event_id=event_id,
                        flow_id=flow_id,
                        module='tracardi.process_engine.rules_engine',
                        class_name="RulesEngine",
                        type="error",
                        message=str(e)
                    )
                    self.console_log.append(console)

                    debug_info = DebugInfo(
                        timestamp=time(),
                        flow=FlowDebugInfo(
                            id=flow_id,
                            error=[ErrorDebugInfo(msg=str(e), file=__file__, line=86)]
                        ),
                        event=Entity(id=event_id)
                    )

                debug_info_by_event_type_and_rule_name[event_type].append({rule_name: debug_info})

        ran_event_types = list(flow_task_store.keys())

        return debug_info_by_event_type_and_rule_name, ran_event_types, self.console_log
        # # Save profile
        #
        # segmentation_info = {"errors": [], "ids": []}
        # try:
        #     save_tasks = []
        #
        #     # Segmentation
        #     if self.profile.operation.needs_update() or self.profile.operation.needs_segmentation():
        #         # Segmentation runs only if profile was updated or flow forced it
        #         async for event_type, segment_id, error in self.profile.segment(
        #                 flow_task_store.keys(),
        #                 load_segment_by_event_type):
        #             # Segmentation triggered
        #             if error:
        #                 segmentation_info['errors'].append(error)
        #             segmentation_info['ids'].append(segment_id)
        #
        #     # Merging, schedule save only if there is an update in flow.
        #
        #     if self.profile.operation.needs_merging() and self.profile.operation.needs_update():
        #
        #         disabled_profiles = await self.profile.merge(limit=2000)
        #         if disabled_profiles is not None:
        #             save_old_profiles_task = asyncio.create_task(
        #                 StorageForBulk(disabled_profiles).index('profile').save())
        #             save_tasks.append(save_old_profiles_task)
        #
        #     # Must be the last operation
        #     if self.profile.operation.needs_update():
        #         save_tasks.append(asyncio.create_task(StorageFor(self.profile).index().save()))
        #
        #     # run save tasks
        #     await asyncio.gather(*save_tasks)
        #
        # except Exception as e:
        #     # this error is a global segmentation error
        #     # todo log it.
        #     print(str(e))
        #     pass
        #
        # finally:
        #     return debug_info_by_event_type_and_rule_name, segmentation_info, console_log

    @staticmethod
    def _mark_profiles_as_merged(profiles, merge_with) -> List[Profile]:
        disabled_profiles = []

        for profile in profiles:
            profile.active = False
            profile.mergedWith = merge_with
            disabled_profiles.append(profile)

        return disabled_profiles

    def _get_merging_keys_and_values(self):
        merge_key_values = self.profile.get_merge_key_values()

        # Add keyword
        merge_key_values = [(f"{field}.keyword", value) for field, value in merge_key_values if value is not None]

        return merge_key_values

    # async def execute(self, source_id) -> Tuple[Dict[str, List[Dict[str, DebugInfo]]], Dict[str, list], list]:
    #
    #     """ Runs rules and flow and saves debug info and console log """
    #
    #     debug_info_by_event_type_and_rule_name, segmentation_info, console_log = await self.invoke(source_id)
    #
    #     # # Save console log
    #     #
    #     # if console_log:
    #     #     await StorageForBulk(console_log).index('console-log').save()
    #     #
    #     # # Collect debug info
    #     #
    #     # record = []
    #     # for debug_info_record in EventDebugRecord.encode(
    #     #         debug_info_by_event_type_and_rule_name):  # type: EventDebugRecord
    #     #     record.append(debug_info_record)
    #     #
    #     # # Save in debug index
    #     # await StorageForBulk(record).index("debug-info").save()
    #     #
    #     # return debug_info_by_event_type_and_rule_name, segmentation_info, console_log



from tracardi.domain.profile import Profile
from tracardi.domain.value_object.bulk_insert_result import BulkInsertResult

from tracardi.service.storage.factory import StorageFor


async def save_profile(profile: Profile, persist_profile: bool = True):
    if persist_profile and profile.operation.new:
        return await StorageFor(profile).index().save()
    else:
        return BulkInsertResult()



from datetime import datetime
from typing import Any
from pydantic import BaseModel


class Metadata(BaseModel):
    timestamp: datetime = None

    def __init__(self, **data: Any):
        super().__init__(**data)
        self.timestamp = datetime.utcnow()


class Console(BaseModel):
    metadata: Metadata = Metadata()
    event_id: str
    flow_id: str
    origin: str
    class_name: str
    module: str
    type: str
    message: str

    # # Persistence
    #
    # def storage(self) -> StorageCrud:
    #     return StorageCrud("console-log", Console, entity=self)
    #
    # @staticmethod
    # def storage_info() -> StorageInfo:
    #     return StorageInfo(
    #         'console-log',
    #         Console
    #     )
#
#
# class ConsoleLog(list):
#
#     def bulk(self) -> CollectionCrud:
#         return CollectionCrud("console-log", self)



import asyncio
from typing import Any, Optional, Coroutine
from uuid import uuid4

from app.api.track.service.tracker import track_event
from app.utils.network import local_ip
from tracardi.domain.payload.event_payload import EventPayload

from tracardi.domain.payload.tracker_payload import TrackerPayload

from tracardi.domain.context import Context
from tracardi.domain.metadata import Metadata
from tracardi.domain.value_object.storage_info import StorageInfo

from tracardi.domain.entity import Entity


class TaskEvent(Entity):
    metadata: Metadata
    type: str
    properties: Optional[dict] = {}

    source: Entity
    session: Entity
    profile: Entity
    context: Context
    options: Optional[dict] = {}

    def to_tracker_payload(self) -> TrackerPayload:
        return TrackerPayload(
            metadata=self.metadata,
            source=self.source,
            session=self.session,
            profile=self.profile,
            context=self.context,
            events=[EventPayload(type=self.type, properties=self.properties, options=self.options)]
        )


class Task(Entity):
    id: str = None
    timestamp: float
    event: TaskEvent
    status: str = 'pending'

    def __init__(self, **data: Any):
        super().__init__(**data)
        if self.id is None:
            self.id = str(uuid4())

    @staticmethod
    def storage_info() -> StorageInfo:
        return StorageInfo(
            'task',
            Task
        )

    def run(self):
        tracker_payload = self.event.to_tracker_payload()

        async def _task():
            try:
                return await track_event(tracker_payload, ip=local_ip), self
            except Exception as e:
                print(str(e))
                self.status = 'error'
                return None, self

        return asyncio.create_task(_task())



class StorageInfo:

    def __init__(self, index, domain_class_ref, exclude=None, exclude_unset=False):
        self.index = index
        self.domain_class_ref = domain_class_ref
        self.exclude = exclude
        self.exclude_unset = exclude_unset




# import base64
# import json
# from typing import List
# from typing import Optional
#
# from tracardi.domain.storage_result import StorageResult
# from tracardi.domain.value_object.storage_info import StorageInfo
# from tracardi.service.storage.collection_crud import CollectionCrud
# from tracardi.domain.named_entity import NamedEntity
# from tracardi.service.storage.crud import StorageCrud
#
#
# class Credential(NamedEntity):
#     description: Optional[str] = ''
#     type: str
#     group: List[str] = ["General"]
#     enabled: bool = True
#     data: dict
#
#
# class CredentialRecord(NamedEntity):
#     description: Optional[str] = ''
#     type: str
#     enabled: bool = True
#     data: str
#     group: List[str] = ["General"]
#
#     @staticmethod
#     def encode(credential: Credential) -> 'CredentialRecord':
#         json_data = json.dumps(credential.data)
#         b64_data = base64.b64encode(json_data.encode('utf-8'))
#
#         return CredentialRecord(
#             id=credential.id,
#             name=credential.name,
#             description=credential.description,
#             type=credential.type,
#             enabled=credential.enabled,
#             group=credential.group,
#             data=b64_data
#         )
#
#     def decode(self) -> Credential:
#         decoded = base64.b64decode(self.data)
#         data = json.loads(decoded)
#
#         return Credential(
#             id=self.id,
#             name=self.name,
#             description=self.description,
#             type=self.type,
#             enabled=self.enabled,
#             group=self.group,
#             data=data
#         )
#
#     # Persistence
#
#     def storage(self) -> StorageCrud:
#         return StorageCrud("credential", CredentialRecord, entity=self)
#
#     @staticmethod
#     def storage_info() -> StorageInfo:
#         return StorageInfo(
#             'credential',
#             CredentialRecord
#         )
#
#
# class CredentialRecords(list):
#
#     def bulk(self) -> CollectionCrud:
#         return CollectionCrud("credential", self)
#
#     @staticmethod
#     def decode(records: StorageResult) -> List[Credential]:
#         return [CredentialRecord.decode(CredentialRecord(**record)) for record in records]



from typing import Optional
from uuid import uuid4
from pydantic import BaseModel
# import tracardi.service.storage.crud as crud
from tracardi.domain.value_object.storage_info import StorageInfo


class Entity(BaseModel):
    id: str

    @staticmethod
    def new() -> 'Entity':
        return Entity(id=str(uuid4()))

    @staticmethod
    def storage_info() -> Optional[StorageInfo]:
        return None



# import logging
# from tracardi.event_server.service.persistence_service import PersistenceService
# from tracardi.service.storage.elastic_storage import ElasticStorage
#
#
# rules_logger = logging.getLogger('Rules')
#
#
# class Rules(list):
#
#     # @staticmethod
#     # def storage() -> PersistenceService:
#     #     return PersistenceService(ElasticStorage(index_key="rule"))



from datetime import datetime
from typing import Optional, Any
from tracardi_plugin_sdk.domain.register import Plugin
from .entity import Entity
from .metadata import Metadata
from .settings import Settings
from .time import Time
from .value_object.storage_info import StorageInfo


class FlowActionPlugin(Entity):

    """
    This object can not be loaded without encoding.
    Load it as FlowActionPluginRecord and then decode.
    """

    metadata: Optional[Metadata]
    plugin: Plugin
    settings: Optional[Settings] = Settings()

    def __init__(self, **data: Any):
        data['metadata'] = Metadata(
            time=Time(
                insert=datetime.utcnow()
            ))
        super().__init__(**data)

    # Persistence

    """
    Do not use load method. This object must be decoded from FlowActionPluginRecord
    """

    @staticmethod
    def storage_info() -> StorageInfo:
        return StorageInfo(
            'action',
            FlowActionPlugin
        )
